/*---------------------------------------------------------------------*/
/* V850 test program: Audio Player                                     */

#include <string.h>
#include "play.h"
#include "comm.h"
#include "iic.h"
#include "monitor.h"
#include "tff.h"
#include "diskio.h"
#include "taskmgr.h"
#include "glcd.h"
#include "gui.h"


/* STA013 control signals */
#define DREQ	P0.4	/* Decoder data request (active high), INTP1 */
#define BITEN	P5.3	/* Decoder chip select (active high) */
#define H		1
#define L		0


/* Transmission FIFO */
static BYTE Fifo[4096];			/* Buffer (must be multiple of 1024) */
static volatile UINT ofs_r, ofs_w, remain;

/* Name tables */
static char AlbumTbl[MAX_ALBMS][13];
static char TrackTbl[MAX_TRKS][13];
static WORD CrcReg;
static DWORD CardTbl[N_EETBL+1];


/* Play status */
static BYTE Cdir, Ctrk;
static BYTE Flags;


extern FATFS fatfs;
extern const BYTE sta013cfg[], sta013start[], sta013stop[];




static
UINT get_key (void)
{
	UINT rv;

	rv = CmdButton;
	if (rv) CmdButton = 0;
	return rv;
}



static
void load_settings (void)
{
	UINT n;
	DWORD d;


	/* Load operating parameters */
	eeprom_read(EE_FLAGS, 1, &Flags);
	eeprom_read(EE_ATT, 1, &Attenuation);
	eeprom_read(EE_CONT, 1, &GuiContrast);

	/* Load card list */
	eeprom_read(EE_TABLE, N_EETBL*4, CardTbl);
	CardTbl[N_EETBL] = 0;
	for (n = 0; n < N_EETBL && CardTbl[n] >> 16 != CrcReg; n++);
	Cdir = (BYTE)CardTbl[n];
	Ctrk = (BYTE)(CardTbl[n] >> 8);
	if (n >= N_EETBL) n = N_EETBL - 1;
	while (n > 0) {
		eeprom_write(EE_TABLE + n * 4, 4, &CardTbl[n - 1]);
		n--;
	}
}



static
void save_settings (void)
{
	CardTbl[0] = ((DWORD)CrcReg << 16) | ((DWORD)Ctrk << 8) | Cdir;
	eeprom_write(EE_TABLE, 4, &CardTbl[0]);
}



/*--------------------------------------------------------------------------*/
/*  File control                                                            */

static
void sort_items (
	char tbl[][13],		/* Address of the name table */
	UINT items			/* Number of items stored in the table */
)
{
	UINT i, j, k;
	char s[13];
	BYTE a, b;

	/* Bubble sort */
	for (i = 0; i + 1 < items; i++) {
		for (j = i + 1; j < items; j++) {
			for (k = 0; k < 13; k++) {
				a = (BYTE)tbl[i][k];
				b = (BYTE)tbl[j][k];
				if (a < b) break;
				if (a > b) {
					memcpy(s, tbl[i], 13);
					memcpy(tbl[i], tbl[j], 13);
					memcpy(tbl[j], s, 13);
					break;
				}
			}
		}
	}
}



static
UINT get_tracks (		/* Number of tracks found */
	char* dirname,		/* Directory to be scanned */
	UINT mode			/* Mode flags: bit0:Use file list, bit1:calcurate crc */
)
{
	UINT ntrks, n;
	DIR dir;
	FILINFO fno;
	FIL fil;
	DWORD crc, b;


//	xprintf("get_album: %s", dirname);
	if (f_opendir(&dir, dirname) != FR_OK) return 0;
	ntrks = 0;
	Fifo[0] = 0;
	while (f_readdir(&dir, &fno) == FR_OK && fno.fname[0]) {
		if (fno.fattrib & AM_DIR) continue;
		if (strstr(fno.fname, ".mp3") || strstr(fno.fname, ".MP3")) {
			if (ntrks < MAX_TRKS) {
				strcpy(TrackTbl[ntrks], fno.fname);
				ntrks++;
			}
			if (mode & 2) {
				crc = CrcReg;
				for (b = 0x80000000; b; b >>= 1) {
					crc <<= 1;
					if (fno.ftime & b) crc |= 1;
					if (crc & 0x10000) crc ^= 0x11021;
				}
				CrcReg = (WORD)crc;
			}
		}
		else if (strstr(fno.fname, ".m3u") || strstr(fno.fname, ".M3U")) {
			strcpy((char*)Fifo, dirname);
			strcat((char*)Fifo, "/");
			strcat((char*)Fifo, fno.fname);
		}
	}

	if (mode & 1) {
		if (f_open(&fil, (char*)Fifo, FA_READ) == FR_OK) {
			ntrks = 0;
			while (ntrks < MAX_TRKS && f_gets((char*)Fifo, sizeof(Fifo), &fil)) {
				if (Fifo[0] != '#' && Fifo[0] > ' ') {
					for (n = 0; n < 13 && Fifo[n] > ' '; n++)
						TrackTbl[ntrks][n] = Fifo[n];
					TrackTbl[ntrks][n] = 0;
					ntrks++;
				}
			}
			f_close(&fil);
		} else {
			sort_items(TrackTbl, ntrks);
		}
	}

//	xprintf(" : %u tracks\n", ntrks);
	return ntrks;
}



static
UINT get_albums (void)
{
	UINT ndirs = 0;
	DIR dir;
	FILINFO fno;


	CrcReg = 0;
	if (f_opendir(&dir, "") == FR_OK) {
		do {
			if (f_readdir(&dir, &fno) != FR_OK || !fno.fname[0]) break;
			if ((fno.fattrib & AM_DIR) && get_tracks(fno.fname, 2)) {
				strcpy(&AlbumTbl[ndirs][0], fno.fname);
				ndirs++;
			}
		} while (ndirs < MAX_ALBMS);
	}
	return ndirs;
}



static
void get_tag (
	FIL* fp
)
{
	UINT n, ver, br;
	INT i;

	/* Get Title and Artist from ID3 v1 tag */
	if (f_lseek(fp, fp->fsize - 128)) return;
	if (f_read(fp, Fifo, 128, &br) || br != 128) return;
	if (!memcmp(&Fifo[0], "TAG", 3)) {
		memcpy(GuiTitle, &Fifo[3], 30); GuiTitle[30] = 0;
		for (i = 29; i >= 0 && (BYTE)GuiTitle[i] <= 0x20; i--) GuiTitle[i] = 0;
		memcpy(GuiArtist, &Fifo[33], 30); GuiArtist[30] = 0;
		for (i = 29; i >= 0 && (BYTE)GuiArtist[i] <= 0x20; i--) GuiArtist[i] = 0;
	}

	/* Get Title and Artist from ID3 v2 tag */
	if (f_lseek(fp, 0)) return;
	if (f_read(fp, &Fifo[0], 10, &br) || br != 10) return;
	if (!memcmp(&Fifo[0], "ID3", 3)) {
		ver = Fifo[3];
		for (;;) {
			if (ver == 2) {
				if (f_read(fp, &Fifo[0], 8, &br) || br != 8) return;
				n = ((DWORD)Fifo[3] << 16) + ((DWORD)Fifo[4] << 8)+ Fifo[5];
			} else {
				if (f_read(fp, &Fifo[0], 10, &br) || br != 10) return;
				n = ((DWORD)Fifo[4] << 24) + ((DWORD)Fifo[5] << 16) + ((DWORD)Fifo[6] << 8) + Fifo[7];
			}
			if (Fifo[0] < 0x41 || Fifo[0] > 0x5A || n > 512) break;
			if (f_read(fp, &Fifo[10], (WORD)n, &br) || br != n) return;
			if (!n || n >= sizeof(GuiTitle) - 1 ) continue;
			if (Fifo[10] != 0) break;
			n--;
			if (!memcmp(Fifo, "TIT2", 4) || !memcmp(&Fifo[0], "TT2", 3)){
				memcpy(GuiTitle, &Fifo[11], n);
				GuiTitle[n] = 0;
			}
			if (!memcmp(&Fifo[0], "TPE1", 4) || !memcmp(&Fifo[0], "TP1", 3)) {
				memcpy(GuiArtist, &Fifo[11], n);
				GuiArtist[n] = 0;
			}

		}
	}
}



static
UINT play_file (	/* Terminated by: any button or 0(eof) */
	char* fname		/* Audio file (mp3) to be played */
)
{
	UINT key, n, ver, br;
	FIL fil;
	BYTE sta013vol[] = {0x46, 0, 0x48, 0, 255, 0};


	if (f_open(&fil, fname, FA_READ|FA_OPEN_EXISTING) != FR_OK) return 0;
	GuiTrkPos = 0;
	GuiTrkSize = fil.fsize;
	strcpy((char*)GuiFile, fname);
	strcpy((char*)GuiTitle, fname);
	GuiArtist[0] = 0;
	get_tag(&fil);
	f_lseek(&fil, 0);
//	xprintf("play_file: %s : %s - %s\n", GuiFile, GuiTitle, GuiArtist);

	GuiCmd = GI_NEW; wakeup_task(2);
	do sleep(0); while (GuiCmd);
	sta013_setregs(sta013start);
	sleep(2);
	sta013vol[1] = sta013vol[3] = Volume;
	sta013_setregs(sta013vol);
	GuiCmd = GI_PLAY; wakeup_task(2);

	ofs_r = ofs_w = remain = 0;			/* Flush FIFO */
	do {
		key = 0;
		PMK1 = 0;		/* Enable STA013 interrupt */
		do {			/* Streaming loop */
			sleep(0);
			/* Store audio data into FIFO when there is >=1024 byte space in the FIFO */
			if (remain <= sizeof(Fifo) - 1024) {
				if (f_read(&fil, &Fifo[ofs_w], 1024, &br) != FR_OK) { key = K_STOP; break; }
				ofs_w = (ofs_w + br) % sizeof(Fifo);
				__DI();
				remain += br;
				if (DREQ == H) PIF1 = 1;	/* Re-trigger INTP1 */
				__EI();
				if (br < 1024) break;	/* EOF? */
				GuiTrkPos = fil.fptr;
			}
			key = get_key();
		} while (!key);
		if (!key) {		/* When reached to EOF */
			do sleep(0); while (remain);	/* Wait for FIFO empty */
		}
		PMK1 = 1;		/* Disable STA013 interrupt */
		if (key == K_PLAY) {	/* Breaking with PLAY button means Pause */
			GuiCmd = GI_PAUSE;
			do {				/* Wait for any button pressed */
				sleep(10);
				key = get_key();
			} while (!key);
			GuiCmd = GI_PLAY;
		}
	} while (key == K_PLAY);

	f_close(&fil);
	GuiCmd = GI_STOP; wakeup_task(2);
	sta013_setregs(sta013stop);
	sleep(2);

	return key;
}



/*--------------------------------------------------------------------------*/
/*  Main task of Audio Player                                               */

void TASK player (void *arg)
{
	char fname[32], buf[2];
	INT ndirs, ntrks, dstat;
	UINT key, key2;


	/* Initialize STA013 */
	sleep(10);
	sta013_setregs(sta013cfg);

	for (;;) {
		dstat = 1;
		do {	/* In STOP state, wait for PLAY button or card insertion */
			sleep(10);
			key = get_key();
			if (disk_status(0) & STA_NODISK) {
				dstat = 0;
			} else {
				if (!dstat) key = K_PLAY;
				dstat = 1;
			}
		} while (!key);
		if (key != K_PLAY) {
			if (key == K_PREVA) {	/* Set loop mode to album repeat */
				Flags |= 0x01; eeprom_write(0, 1, &Flags);
			}
			if (key == K_PREV) {	/* Set loop mode to all repeat */
				Flags &= 0xFE; eeprom_write(0, 1, &Flags);
			}
			continue;
		}

		/* Create directory (album) list */
		f_mount(0, &fatfs);
		ndirs = get_albums();
		if (!ndirs) continue;
		sort_items(AlbumTbl, ndirs);
		load_settings();
		GuiCmd = GI_INIT; wakeup_task(2);
		do sleep(0); while(GuiCmd);

		key = 0;
		for (;;) {
			if (Cdir >= ndirs) Cdir = 0;
			ntrks = get_tracks(AlbumTbl[Cdir], 1);	/* Get tracks of current album */
			if (!ntrks) break;
			GuiNumTrks = ntrks;
			if (key == K_PREV) {
				Ctrk = ntrks - 1;
			} else if (Ctrk >= ntrks) {
				Ctrk = 0;
			}
			for (;;) {
				strcpy(fname, AlbumTbl[Cdir]);	/* Play a track */
				strcat(fname, "/");
				strcat(fname, TrackTbl[Ctrk]);
				GuiCurrTrk = Ctrk + 1;
				key = play_file(fname);
				if (key == K_STOP) break;
				if (key == K_PREVA) {	/* PREVA */
					if (--Cdir >= ndirs) Cdir = ndirs - 1;	/* Goto top of previous album */
					Ctrk = 0;
					break;
				}
				if (key == K_NEXTA) {	/* NEXTA */
					Cdir++; Ctrk = 0;					/* Goto top of next album */
					break;
				}
				if (key == K_PREV) {	/* PREV */
					if (--Ctrk >= ntrks) {
						if ((Flags & 1) && (--Cdir >= ndirs)) Cdir = ndirs - 1;
						break;
					}
				} else {				/* EOF or NEXT */
					if (++Ctrk >= ntrks) {
						if (Flags & 1) Cdir++;
						Ctrk = 0;
						break;
					}
				}
			}
			if (key == K_STOP) break;
		}
		/* Save current track and album */
		save_settings();
	}
}



/*--------------------------------------------------*/
/* INTP1 interrupt service routine triggerd by      */
/* rising edge of DREQ                              */
/*--------------------------------------------------*/

#pragma interrupt INTP1 ISR_intp1
__multi_interrupt
void ISR_intp1 (void)
{
	UINT rptr, cnt;
	BYTE d;


	__EI();

	cnt = remain;
	if (cnt > 0 && DREQ == H) {
		rptr = ofs_r;
		BITEN = H;
		do {
			d = Fifo[rptr];
			rptr = (rptr + 1) % sizeof(Fifo);
			do; while (CB2TSF);
			CB2TXL = d;
		} while (--cnt > 0 && DREQ == H);
		do; while (CB2TSF);
		BITEN = L;
		ofs_r = rptr;
		remain = cnt;
	}

	__DI();
}


